# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""WSGI Resource definition for Glare. Defines Glare API and serialization/
deserialization of incoming requests."""

import json
import jsonpatch
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
import six
from six.moves import http_client
import six.moves.urllib.parse as urlparse

from glare.api.v1 import api_versioning
from glare.common import exception as exc
from glare.common import wsgi
from glare import engine
from glare.i18n import _, _LI

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

list_configs = [
    cfg.IntOpt('default_api_limit', default=25,
               help=_('Default value for the number of items returned by a '
                      'request if not specified explicitly in the request')),
    cfg.IntOpt('max_api_limit', default=1000,
               help=_('Maximum permissible number of items that could be '
                      'returned by a request')),
]

CONF.register_opts(list_configs)

supported_versions = api_versioning.VersionedResource.supported_versions


class RequestDeserializer(api_versioning.VersionedResource,
                          wsgi.JSONRequestDeserializer):
    """Glare deserializer for incoming webop Requests.
    Deserializer converts incoming request into bunch of python primitives.
    So other components doesn't work with requests at all. Deserializer also
    executes primary API validation without any knowledge about Artifact
    structure.
    """

    @staticmethod
    def _get_content_type(req, expected=None):
        """Determine content type of the request body."""
        if "Content-Type" not in req.headers:
            msg = _("Content-Type must be specified.")
            LOG.error(msg)
            raise exc.BadRequest(msg)

        content_type = req.content_type
        if expected is not None and content_type not in expected:
            msg = (_('Invalid content type: %(ct)s. Expected: %(exp)s') %
                   {'ct': content_type, 'exp': ', '.join(expected)})
            raise exc.UnsupportedMediaType(message=msg)

        return content_type

    def _get_request_body(self, req):
        return self.from_json(req.body)

    @supported_versions(min_ver='1.0')
    def create(self, req):
        self._get_content_type(req, expected=['application/json'])
        body = self._get_request_body(req)
        if not isinstance(body, dict):
            msg = _("Dictionary expected as body value. Got %s.") % type(body)
            raise exc.BadRequest(msg)
        return {'values': body}

    @supported_versions(min_ver='1.0')
    def list(self, req):
        params = req.params.copy()
        marker = params.pop('marker', None)
        query_params = {}
        # step 1 - apply marker to query if exists
        if marker is not None:
            query_params['marker'] = marker

        # step 2 - apply limit (if exists OR setup default limit)
        limit = params.pop('limit', CONF.default_api_limit)
        try:
            limit = int(limit)
        except ValueError:
            msg = _("Limit param must be an integer.")
            raise exc.BadRequest(message=msg)
        if limit < 0:
            msg = _("Limit param must be positive.")
            raise exc.BadRequest(message=msg)
        query_params['limit'] = min(CONF.max_api_limit, limit)

        # step 3 - parse sort parameters
        if 'sort' in params:
            sort = []
            for sort_param in params.pop('sort').strip().split(','):
                key, _sep, direction = sort_param.partition(':')
                if direction and direction not in ('asc', 'desc'):
                    raise exc.BadRequest('Sort direction must be one of '
                                         '["asc", "desc"]. Got %s direction'
                                         % direction)
                sort.append((key, direction or 'desc'))
            query_params['sort'] = sort

        # step 4 - parse filter parameters
        filters = []
        for fname, fval in six.iteritems(params):
            if fname == 'version' and fval == 'latest':
                query_params['latest'] = True
            else:
                filters.append((fname, fval))

        query_params['filters'] = filters
        return query_params

    @supported_versions(min_ver='1.0')
    def update(self, req):
        self._get_content_type(
            req, expected=['application/json-patch+json'])
        body = self._get_request_body(req)
        patch = jsonpatch.JsonPatch(body)
        try:
            # Initially patch object doesn't validate input. It's only checked
            # we call get operation on each method
            tuple(map(patch._get_operation, patch.patch))
        except (jsonpatch.InvalidJsonPatch, TypeError):
            msg = _("Json Patch body is malformed")
            raise exc.BadRequest(msg)
        for patch_item in body:
            if patch_item['path'] == '/tags':
                msg = _("Cannot modify artifact tags with PATCH "
                        "request. Use special Tag API for that.")
                raise exc.BadRequest(msg)
        return {'patch': patch}

    def _deserialize_blob(self, req):
        content_type = self._get_content_type(req)
        if content_type == ('application/vnd+openstack.glare-custom-location'
                            '+json'):
            data = self._get_request_body(req)
            if 'url' not in data:
                msg = _("url is required when specifying external location. "
                        "Cannot find url in body: %s") % str(data)
                raise exc.BadRequest(msg)
        else:
            data = req.body_file
        return {'data': data, 'content_type': content_type}

    @supported_versions(min_ver='1.0')
    def upload_blob(self, req):
        return self._deserialize_blob(req)

    @supported_versions(min_ver='1.0')
    def upload_blob_dict(self, req):
        return self._deserialize_blob(req)

    @supported_versions(min_ver='1.0')
    def set_tags(self, req):
        self._get_content_type(req, expected=['application/json'])
        body = self._get_request_body(req)

        if 'tags' not in body:
            msg = _("Tag list must be in the body of request.")
            raise exc.BadRequest(msg)

        return {'tag_list': body['tags']}


def log_request_progress(f):
    def log_decorator(self, req, *args, **kwargs):
        LOG.debug("Request %(request_id)s for %(api_method)s successfully "
                  "deserialized. Pass request parameters to Engine",
                  {'request_id': req.context.request_id,
                   'api_method': f.__name__})
        result = f(self, req, *args, **kwargs)
        LOG.info(_LI(
            "Request %(request_id)s for artifact %(api_method)s "
            "successfully executed."), {'request_id': req.context.request_id,
                                        'api_method': f.__name__})
        return result
    return log_decorator


class ArtifactsController(api_versioning.VersionedResource):
    """API controller for Glare Artifacts.
    Artifact Controller prepares incoming data for Glare Engine and redirects
    data to appropriate engine method (so only controller is working with
    Engine. Once the data returned from Engine Controller returns data
    in appropriate format for Response Serializer.
    """

    def __init__(self):
        self.engine = engine.Engine()

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def list_type_schemas(self, req):
        type_schemas = self.engine.list_type_schemas(req.context)
        return type_schemas

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def show_type_schema(self, req, type_name):
        type_schema = self.engine.show_type_schema(req.context, type_name)
        return {type_name: type_schema}

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def create(self, req, type_name, values):
        """Create artifact record in Glare.

        :param req: User request
        :param type_name: Artifact type name
        :param values: dict with artifact fields {field_name: field_value}
        :return definition of created artifact
        """
        return self.engine.create(req.context, type_name, values)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def update(self, req, type_name, artifact_id, patch):
        """Update artifact record in Glare.

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of artifact to update
        :param patch: json patch with artifact changes
        :return definition of updated artifact
        """
        return self.engine.update(req.context, type_name, artifact_id, patch)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def delete(self, req, type_name, artifact_id):
        """Delete artifact from Glare

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of artifact to delete
        """
        return self.engine.delete(req.context, type_name, artifact_id)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def show(self, req, type_name, artifact_id):
        """Show detailed artifact info

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of artifact to show
        :return: definition of requested artifact
        """
        return self.engine.get(req.context, type_name, artifact_id)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def list(self, req, type_name, filters, marker=None, limit=None,
             sort=None, latest=False):
        """List available artifacts

        :param req: User request
        :param type_name: Artifact type name
        :param filters: filters that need to be applied to artifact
        :param marker: the artifact that considered as begin of the list
        so all artifacts before marker (including marker itself) will not be
        added to artifact list
        :param limit: maximum number of items in list
        :param sort: sorting options
        :param latest: flag that indicates, that only artifacts with highest
        versions should be returned in output
        :return: list of artifacts
        """
        artifacts = self.engine.list(req.context, type_name, filters, marker,
                                     limit, sort, latest)
        result = {'artifacts': artifacts,
                  'type_name': type_name}
        if len(artifacts) != 0 and len(artifacts) == limit:
            result['next_marker'] = artifacts[-1]['id']
        return result

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def upload_blob(self, req, type_name, artifact_id, field_name, data,
                    content_type):
        """Upload blob into Glare repo

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of Artifact to reactivate
        :param field_name: name of blob field in artifact
        :param data: Artifact payload
        :param content_type: data content-type
        """
        if content_type == ('application/vnd+openstack.glare-custom-location'
                            '+json'):
            url = data.pop('url')
            return self.engine.add_blob_location(
                req.context, type_name, artifact_id, field_name, url, data)
        else:
            return self.engine.upload_blob(req.context, type_name, artifact_id,
                                           field_name, data, content_type)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def upload_blob_dict(self, req, type_name, artifact_id, field_name, data,
                         blob_key, content_type):
        """Upload blob into Glare repo

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of Artifact to reactivate
        :param field_name: name of blob field in artifact
        :param data: Artifact payload
        :param content_type: data content-type
        :param blob_key: blob key in dict
        """
        if content_type == ('application/vnd+openstack.glare-custom-location'
                            '+json'):
            url = data.pop('url')
            return self.engine.add_blob_dict_location(
                req.context, type_name, artifact_id,
                field_name, blob_key, url, data)
        else:
            return self.engine.upload_blob_dict(
                req.context, type_name, artifact_id,
                field_name, blob_key, data, content_type)

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def download_blob(self, req, type_name, artifact_id, field_name):
        """Download blob data from Artifact

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of Artifact to reactivate
        :param field_name: name of blob field in artifact
        :return: iterator that returns blob data
        """
        data, meta = self.engine.download_blob(req.context, type_name,
                                               artifact_id, field_name)
        result = {'data': data, 'meta': meta}
        return result

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def download_blob_dict(self, req, type_name, artifact_id,
                           field_name, blob_key):
        """Download blob data from Artifact

        :param req: User request
        :param type_name: Artifact type name
        :param artifact_id: id of Artifact to reactivate
        :param field_name: name of blob field in artifact
        :param blob_key: name of Dict of blobs (optional)
        :return: iterator that returns blob data
        """
        data, meta = self.engine.download_blob_dict(
            req.context, type_name, artifact_id, field_name, blob_key)
        result = {'data': data, 'meta': meta}
        return result

    @staticmethod
    def _tag_body_resp(af):
        return {'tags': af['tags']}

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def get_tags(self, req, type_name, artifact_id):
        return self._tag_body_resp(self.engine.get(
            req.context, type_name, artifact_id))

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def set_tags(self, req, type_name, artifact_id, tag_list):
        patch = [{'op': 'replace', 'path': '/tags', 'value': tag_list}]
        patch = jsonpatch.JsonPatch(patch)
        return self._tag_body_resp(self.engine.update(
            req.context, type_name, artifact_id, patch))

    @supported_versions(min_ver='1.0')
    @log_request_progress
    def delete_tags(self, req, type_name, artifact_id):
        patch = [{'op': 'replace', 'path': '/tags', 'value': []}]
        patch = jsonpatch.JsonPatch(patch)
        self.engine.update(req.context, type_name, artifact_id, patch)


class ResponseSerializer(api_versioning.VersionedResource,
                         wsgi.JSONResponseSerializer):
    """Glare Response Serializer converts data received from Glare Engine
    (it consists from plain data types - dict, int, string, file descriptors,
    etc) to WSGI Requests. It also specifies proper response status and
    content type as specified by API design.
    """

    @staticmethod
    def _prepare_json_response(response, result,
                               content_type='application/json'):
        body = json.dumps(result, ensure_ascii=False)
        response.unicode_body = six.text_type(body)
        response.content_type = content_type

    def list_type_schemas(self, response, type_schemas):
        self._prepare_json_response(response,
                                    {'schemas': type_schemas},
                                    content_type='application/schema+json')

    def show_type_schema(self, response, type_schema):
        self._prepare_json_response(response,
                                    {'schemas': type_schema},
                                    content_type='application/schema+json')

    @supported_versions(min_ver='1.0')
    def list_schemas(self, response, type_list):
        self._prepare_json_response(response, {'types': type_list})

    @supported_versions(min_ver='1.0')
    def create(self, response, artifact):
        self._prepare_json_response(response, artifact)
        response.status_int = http_client.CREATED

    @supported_versions(min_ver='1.0')
    def show(self, response, artifact):
        self._prepare_json_response(response, artifact)

    @supported_versions(min_ver='1.0')
    def update(self, response, artifact):
        self._prepare_json_response(response, artifact)

    @supported_versions(min_ver='1.0')
    def list(self, response, af_list):
        params = dict(response.request.params)
        params.pop('marker', None)

        encode_params = {}
        for key, value in six.iteritems(params):
            encode_params[key] = encodeutils.safe_encode(value)
        query = urlparse.urlencode(encode_params)

        type_name = af_list['type_name']
        body = {
            type_name: af_list['artifacts'],
            'first': '/artifacts/%s' % type_name,
            'schema': '/schemas/%s' % type_name,
        }
        if query:
            body['first'] = '%s?%s' % (body['first'], query)
        if 'next_marker' in af_list:
            params['marker'] = af_list['next_marker']
            next_query = urlparse.urlencode(params)
            body['next'] = '/artifacts/%s?%s' % (type_name, next_query)
        response.unicode_body = six.text_type(json.dumps(body,
                                                         ensure_ascii=False))
        response.content_type = 'application/json'

    @supported_versions(min_ver='1.0')
    def delete(self, response, result):
        response.status_int = http_client.NO_CONTENT

    @supported_versions(min_ver='1.0')
    def upload_blob(self, response, artifact):
        self._prepare_json_response(response, artifact)

    @staticmethod
    def _serialize_blob(response, result):
        data, meta = result['data'], result['meta']
        response.headers['Content-Type'] = meta['content_type']
        response.headers['Content-MD5'] = meta['md5']
        response.headers['X-Openstack-Glare-Content-SHA1'] = meta['sha1']
        response.headers['X-Openstack-Glare-Content-SHA256'] = meta['sha256']
        response.app_iter = iter(data)
        response.content_length = str(meta['size'])

    @staticmethod
    def _serialize_location(response, result):
        data, meta = result['data'], result['meta']
        response.headers['Content-MD5'] = meta['md5']
        response.headers['X-Openstack-Glare-Content-SHA1'] = meta['sha1']
        response.headers['X-Openstack-Glare-Content-SHA256'] = meta['sha256']
        response.location = data['url']
        response.content_type = 'application/json'
        response.status = http_client.MOVED_PERMANENTLY
        response.content_length = 0

    @supported_versions(min_ver='1.0')
    def download_blob(self, response, result):
        external = result['meta']['external']
        if external:
            self._serialize_location(response, result)
        else:
            self._serialize_blob(response, result)

    @supported_versions(min_ver='1.0')
    def download_blob_dict(self, response, result):
        external = result['meta']['external']
        if external:
            self._serialize_location(response, result)
        else:
            self._serialize_blob(response, result)

    @supported_versions(min_ver='1.0')
    def delete_tags(self, response, result):
        response.status_int = http_client.NO_CONTENT


def create_resource():
    """Artifact resource factory method"""
    deserializer = RequestDeserializer()
    serializer = ResponseSerializer()
    controller = ArtifactsController()
    return wsgi.Resource(controller, deserializer, serializer)



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields
import six
import six.moves.urllib.request as urlrequest

from glare.common import exception
from glare.common import store_api
from glare.common import utils
from glare.db import artifact_api
from glare import locking
from glare.i18n import _, _LI
from glare.objects.meta import attribute
from glare.objects.meta import fields as glare_fields
from glare.objects.meta import validators

artifact_opts = [
    cfg.BoolOpt('delayed_blob_delete', default=False,
                help=_("Defines if blob must be deleted immediately "
                       "or just marked as deleted so it can be cleaned by some"
                       "other tool in background.")),
]

CONF = cfg.CONF
CONF.register_opts(artifact_opts)

LOG = logging.getLogger(__name__)


class classproperty(property):
    """Special decorator that creates class properties"""

    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class BaseArtifact(base.VersionedObject):
    """BaseArtifact is a central place in Glare. It execute Glare business
    logic operations and checks in like:
    1) Check if artifact satisfies all requirements and can be activated
    2) Check that artifact is not deactivated and download blobs
    ...
    BaseArtifact interacts with database and saves/request artifact info
    from specified database API. Base Artifact is an abstract class so
    all concrete classes must be inherited from that class. Concrete classes
    must define custom fields in addition to BaseArtifact fields and db_api
    that must be used for interaction with database.
    """

    OBJ_PROJECT_NAMESPACE = 'glare'

    DEFAULT_ARTIFACT_VERSION = '0.0.0'

    STATUS = glare_fields.ArtifactStatusField

    Field = attribute.Attribute.init
    DictField = attribute.DictAttribute.init
    ListField = attribute.ListAttribute.init
    Blob = attribute.BlobAttribute.init

    fields = {
        'id': Field(fields.StringField, system=True,
                    validators=[validators.UUID()], nullable=False,
                    sortable=True, description="Artifact UUID."),
        'name': Field(fields.StringField, required_on_activate=False,
                      nullable=False, sortable=True,
                      description="Artifact Name."),
        'owner': Field(fields.StringField, system=True,
                       required_on_activate=False, nullable=False,
                       sortable=True, description="ID of user/tenant who "
                                                  "uploaded artifact."),
        'status': Field(glare_fields.ArtifactStatusField,
                        default=glare_fields.ArtifactStatusField.DRAFTED,
                        nullable=False, sortable=True,
                        description="Artifact status."),
        'created_at': Field(fields.DateTimeField, system=True,
                            filter_ops=attribute.FILTERS,
                            nullable=False, sortable=True,
                            description="Datetime when artifact has "
                                        "been created."),
        'updated_at': Field(fields.DateTimeField, system=True,
                            filter_ops=attribute.FILTERS,
                            nullable=False, sortable=True,
                            description="Datetime when artifact has "
                                        "been updated last time."),
        'activated_at': Field(fields.DateTimeField, system=True,
                              filter_ops=attribute.FILTERS,
                              required_on_activate=False, sortable=True,
                              description="Datetime when artifact has became "
                                          "active."),
        'description': Field(fields.StringField, mutable=True,
                             required_on_activate=False, default="",
                             validators=[validators.MaxStrLen(4096)],
                             filter_ops=[],
                             description="Artifact description."),
        'tags': ListField(fields.String, mutable=True,
                          required_on_activate=False,
                          # tags are filtered without any operators
                          filter_ops=[],
                          element_validators=[validators.ForbiddenChars(
                              [',', '/'])],
                          description="List of tags added to Artifact."),
        'metadata': DictField(fields.String, required_on_activate=False,
                              element_validators=[validators.MinStrLen(1)],
                              filter_ops=(attribute.FILTER_EQ,
                                          attribute.FILTER_NEQ),
                              description="Key-value dict with useful "
                                          "information about an artifact."),
        'visibility': Field(fields.StringField, default='private',
                            nullable=False, filter_ops=(attribute.FILTER_EQ,),
                            sortable=True,
                            description="Artifact visibility that defines "
                                        "if artifact can be available to "
                                        "other users."),
        'version': Field(glare_fields.VersionField, required_on_activate=False,
                         default=DEFAULT_ARTIFACT_VERSION,
                         filter_ops=attribute.FILTERS, nullable=False,
                         sortable=True, validators=[validators.Version()],
                         description="Artifact version(semver)."),
        'provided_by': DictField(fields.String,
                                 validators=[
                                     validators.AllowedDictKeys(
                                         ("name", "href", "company")),
                                     validators.RequiredDictKeys(
                                         ("name", "href", "company"))
                                 ],
                                 default=None,
                                 required_on_activate=False,
                                 description="Info about artifact authors."),
        'supported_by': DictField(fields.String,
                                  validators=[
                                      validators.RequiredDictKeys(("name",))
                                  ],
                                  default=None,
                                  required_on_activate=False,
                                  description="Info about persons who "
                                              "responsible for artifact "
                                              "support"),
        'release': ListField(fields.String,
                             validators=[validators.Unique()],
                             required_on_activate=False,
                             description="Target Openstack release "
                                         "for artifact. It is usually the same"
                                         " when artifact was uploaded."),
        'icon': Blob(required_on_activate=False,
                     description="Artifact icon."),
        'license': Field(fields.StringField,
                         required_on_activate=False,
                         description="Artifact license type."),
        'license_url': Field(fields.StringField,
                             required_on_activate=False,
                             description="URL to artifact license."),
    }

    @classmethod
    def is_blob(cls, field_name):
        """Helper to check that field is blob

        :param field_name: name of field
        :return: True if field is a blob, False otherwise
        """
        return isinstance(cls.fields.get(field_name), glare_fields.BlobField)

    @classmethod
    def is_blob_dict(cls, field_name):
        """Helper to check that field is blob dict

        :param field_name: name of field
        :return: True if field is a blob dict, False otherwise
        """
        return (isinstance(cls.fields.get(field_name), glare_fields.Dict) and
                cls.fields[field_name].element_type ==
                glare_fields.BlobFieldType)

    @classmethod
    def _init_artifact(cls, context, values):
        """Initialize an empty versioned object with values

        Initialize vo object with default values and values specified by user.
        Also reset all changes for initialized object so user of the method
        can track own changes.

        :param context: user context
        :param values: values needs to be set
        :return: artifact with initialized values
        """
        af = cls(context)
        # setup default values for all non specified attributes
        default_attrs = []
        for attr in af.fields:
            if attr not in values:
                default_attrs.append(attr)
        if default_attrs:
            af.obj_set_defaults(*default_attrs)
        for name, value in six.iteritems(values):
            setattr(af, name, value)
        return af

    @classmethod
    def get_type_name(cls):
        """Return type name that allows to find Artifact Type in Glare

        Type name allows to find Artifact Type definition in Glare registry
        so Engine can instantiate Artifacts. Artifact also becomes available
        with artifact type in Glare API.
        For example, when get_type_name returns 'my_artifact' then
        users can list artifacts by GET <host_name>/v1/artifacts/my_artifact.
        This type name is also used in glare configuration when turning on/off
        specific Artifact Types.
        :return: string that identifies current Artifact Type.
        """
        raise NotImplementedError()

    _DB_API = None

    @classmethod
    def init_db_api(cls):
        """Provide initialized db api to interact with artifact database.

        To interact with database each artifact type must provide an api
        to execute db operations with artifacts.
        :return: subtype of glare.db.api.BaseDBAPI
        """
        return artifact_api.ArtifactAPI(cls)

    @classproperty
    def db_api(cls):
        """Return current database API"""
        if cls._DB_API is None:
            cls._DB_API = cls.init_db_api()
        return cls._DB_API

    lock_engine = locking.LockEngine(artifact_api.ArtifactLockApi())

    @classmethod
    def _lock_version(cls, context, values):
        """Calculate version scope for new artifact

        :param values: af values
        :return: string that identifies af version or None
        """
        name = values.get('name')
        version = values.get('version', cls.DEFAULT_ARTIFACT_VERSION)
        scope_id = None
        if name:
            scope_id = "%s:%s" % (name, str(version))

        return cls.lock_engine.acquire(context, scope_id)

    @classmethod
    def _lock_updated_version(cls, af, updates):
        name = updates.get('name', af.name)
        version = updates.get('version', af.version)
        visibility = updates.get('visibility', af.visibility)
        scope_id = None
        if (name, version, visibility) != (af.name, af.version, af.visibility):
            # no version change == no lock for version
            scope_id = "%s:%s" % (name, str(version))
            if visibility != 'public':
                scope_id += ':%s' % str(af.obj_context.tenant)

        return cls.lock_engine.acquire(af.obj_context, scope_id)

    @classmethod
    def create(cls, context, values):
        """Create new Artifact in Glare repo

        :param context: user context
        :param values: Dict with specified artifact properties
        :return: definition of create Artifact
        """
        if context.tenant is None or context.read_only:
            msg = _("It's forbidden to anonymous users to create artifacts.")
            raise exception.Forbidden(msg)
        else:
            with cls._lock_version(context, values):
                ver = values.setdefault(
                    'version', cls.DEFAULT_ARTIFACT_VERSION)
                cls._validate_versioning(context, values.get('name'), ver)
                # validate other values
                cls._validate_input_values(context, values)
                # validate visibility
                if 'visibility' in values:
                    msg = _("visibility is not allowed in a request "
                            "for artifact create.")
                    raise exception.BadRequest(msg)
                values['id'] = str(uuid.uuid4())
                values['owner'] = context.tenant
                values['created_at'] = timeutils.utcnow()
                values['updated_at'] = values['created_at']
                af = cls._init_artifact(context, values)
                LOG.info(_LI("Parameters validation for artifact creation "
                             "passed for request %s."), context.request_id)
                af_vals = cls.db_api.create(context,
                                            af.obj_changes_to_primitive())
                return cls._init_artifact(context, af_vals)

    @classmethod
    def _validate_versioning(cls, context, name, version, is_public=False):
        if version is not None and name not in (None, ""):
            filters = [('name', name), ('version', version),
                       ('status', 'neq:deleted')]
            if is_public is False:
                filters.extend([('owner', context.tenant),
                                ('visibility', 'private')])
            else:
                filters.extend([('visibility', 'public')])
            if len(cls.list(context, filters)) > 0:
                msg = _("Artifact with this name and version is already "
                        "exists for this owner.")
                raise exception.Conflict(msg)
        else:
            msg = _("Cannot set artifact version without name and version.")
            raise exception.BadRequest(msg)

    @classmethod
    def _validate_input_values(cls, context, values):
        # validate that we are not specifying any system attribute
        # and that we do not upload blobs or add locations here
        for field_name in values:
            if field_name in cls.fields:
                if cls.fields[field_name].system is True:
                    msg = _("Cannot specify system property %s. It is not "
                            "available for modifying by users.") % field_name
                    raise exception.Forbidden(msg)
                elif cls.is_blob(field_name) or cls.is_blob_dict(field_name):
                    msg = _("Cannot add blob %s with this request. "
                            "Use special Blob API for that.") % field_name
                    raise exception.BadRequest(msg)
            else:
                msg = (_("Cannot add non-existing property %s to artifact. ")
                       % field_name)
                raise exception.BadRequest(msg)

    @classmethod
    def _validate_update_allowed(cls, context, af, field_names):
        """Validate if fields can be updated in artifact

        :param context:
        :param af:
        :param field_names:
        :return:
        """
        if af.status not in (cls.STATUS.ACTIVE, cls.STATUS.DRAFTED):
            msg = _("Forbidden to change attributes "
                    "if artifact not active or drafted.")
            raise exception.Forbidden(message=msg)

        for field_name in field_names:
            field = cls.fields[field_name]
            if field.system is True:
                msg = _("Cannot specify system property %s. It is not "
                        "available for modifying by users.") % field_name
                raise exception.Forbidden(msg)
            if af.status == cls.STATUS.ACTIVE and not field.mutable:
                msg = (_("Forbidden to change property '%s' after activation.")
                       % field_name)
                raise exception.Forbidden(message=msg)

    @classmethod
    def update(cls, context, af, values):
        """Update Artifact in Glare repo

        :param context: user Context
        :param af: current definition of Artifact in Glare
        :param values: list of changes for artifact
        :return: definition of updated Artifact
        """
        # reset all changes of artifact to reuse them after update
        af.obj_reset_changes()
        with cls._lock_updated_version(af, values):
            # validate version
            if 'name' in values or 'version' in values:
                new_name = values.get('name') or af.name
                new_version = values.get('version') or af.version
                cls._validate_versioning(context, new_name, new_version)

            # validate other values
            cls._validate_update_allowed(context, af, list(values))
            cls._validate_input_values(context, values)
            # apply values to the artifact. if all changes applied then update
            # values in db or raise an exception in other case.
            for key, value in six.iteritems(values):
                setattr(af, key, value)

            LOG.info(_LI("Parameters validation for artifact %(artifact)s "
                         "update passed for request %(request)s."),
                     {'artifact': af.id, 'request': context.request_id})
            updated_af = cls.db_api.update(
                context, af.id, af.obj_changes_to_primitive())
            return cls._init_artifact(context, updated_af)

    @classmethod
    def get_action_for_updates(cls, context, artifact, updates, registry):
        """The method defines how to detect appropriate action based on update

        Validate request for update and determine if it is request for action.
        Also do a validation for request for action if it is an action.

        :return: action reference for updates dict
        """
        action = cls.update
        if 'visibility' in updates:
            # validate publish action format
            action = cls.publish
        elif 'status' in updates:
            status = updates['status']
            if status == cls.STATUS.DEACTIVATED:
                action = cls.deactivate
            elif status == cls.STATUS.ACTIVE:
                if artifact.status == artifact.STATUS.DEACTIVATED:
                    action = cls.reactivate
                else:
                    action = cls.activate

        # check updates for links and validate them
        try:
            for key, value in six.iteritems(updates):
                if cls.fields.get(key) is glare_fields.Link \
                        and value is not None:
                    # check format
                    glare_fields.LinkFieldType.coerce(None, key, value)
                    # check containment
                    if glare_fields.LinkFieldType.is_external(value):
                        # validate external link
                        cls._validate_external_link(value)
                    else:
                        type_name = (glare_fields.LinkFieldType.
                                     get_type_name(value))
                        af_type = registry.get_artifact_type(type_name)
                        cls._validate_soft_link(context, value, af_type)
        except Exception as e:
            msg = (_("Bad link in artifact %(af)s: %(msg)s")
                   % {"af": artifact.id, "msg": str(e)})
            raise exception.BadRequest(msg)

        LOG.debug("Action %(action)s defined to updates %(updates)s.",
                  {'action': action.__name__, 'updates': updates})

        return action

    @classmethod
    def _validate_external_link(cls, link):
        with urlrequest.urlopen(link) as data:
            data.read(1)

    @classmethod
    def _validate_soft_link(cls, context, link, af_type):
        af_id = link.split('/')[3]
        af_type.get(context, af_id)

    @classmethod
    def get(cls, context, artifact_id):
        """Return Artifact from Glare repo

        :param context: user context
        :param artifact_id: id of requested Artifact
        :return: Artifact definition
        """
        af = cls.db_api.get(context, artifact_id)
        return cls._init_artifact(context, af)

    @classmethod
    def _get_field_type(cls, obj):
        if isinstance(obj, fields.IntegerField) or obj is fields.Integer:
            return 'int'
        elif isinstance(obj, fields.FloatField) or obj is fields.Float:
            return 'numeric'
        elif isinstance(obj, fields.FlexibleBooleanField) or \
                obj is fields.FlexibleBoolean:
            return 'bool'
        return 'string'

    @classmethod
    def _parse_sort_values(cls, sort):
        new_sort = []
        for key, direction in sort:
            if key not in cls.fields:
                msg = _("The field %s doesn't exist.") % key
                raise exception.BadRequest(msg)
            # check if field can be sorted
            if not cls.fields[key].sortable:
                msg = _("The field %s is not sortable.") % key
                raise exception.BadRequest(msg)
            new_sort.append((key, direction, cls._get_field_type(
                cls.fields.get(key))))
        return new_sort

    @classmethod
    def _validate_filter_name(cls, filter_name):
        if cls.fields.get(filter_name) is None:
            msg = _("Unable filter '%s'") % filter_name
            raise exception.BadRequest(msg)

    @classmethod
    def _validate_filter_ops(cls, filter_name, op):
        field = cls.fields.get(filter_name)
        if op not in field.filter_ops:
            msg = (_("Unsupported filter type '%s(key)'."
                     "The following filters are supported "
                     "%(filters)s") % {
                'key': op, 'filters': str(field.filter_ops)})
            raise exception.BadRequest(message=msg)

    @classmethod
    def _parse_filter_values(cls, filters):
        # input format for filters is list of tuples:
        # (filter_name, filter_value)
        # output format for filters is list of tuples:
        # (field_name, key_name, op, field_type, value)
        new_filters = []
        for filter_name, filter_value in filters:
            if filter_name in ('tags-any', 'tags'):
                if ':' in filter_value:
                    msg = _("Tags are filtered without operator")
                    raise exception.BadRequest(msg)
                new_filters.append(
                    (filter_name, None, None, None, filter_value))
                continue

            key_name = None
            if '.' in filter_name:
                filter_name, key_name = filter_name.split('.', 1)
                if not isinstance(cls.fields.get(filter_name),
                                  glare_fields.Dict):
                    msg = _("Field %s is not Dict") % filter_name
                    raise exception.BadRequest(msg)

            cls._validate_filter_name(filter_name)
            field_type = cls.fields.get(filter_name)

            if isinstance(field_type, glare_fields.List) or isinstance(
                    field_type, glare_fields.Dict) and key_name is not None:
                field_type = field_type.element_type
            try:
                if isinstance(field_type, glare_fields.Dict):
                    new_filters.append((
                        filter_name, filter_value, None, None, None))
                else:
                    op, val = utils.split_filter_op(filter_value)
                    cls._validate_filter_ops(filter_name, op)
                    if op == 'in':
                        value = [field_type.coerce(cls(), filter_name, value)
                                 for value in
                                 utils.split_filter_value_for_quotes(val)]
                    else:
                        value = field_type.coerce(cls(), filter_name, val)
                    new_filters.append(
                        (filter_name, key_name, op,
                         cls._get_field_type(field_type), value))
            except ValueError:
                msg = _("Invalid filter value: %s") % str(val)
                raise exception.BadRequest(msg)

        return new_filters

    @classmethod
    def list(cls, context, filters=None, marker=None, limit=None,
             sort=None, latest=False):
        """List all available Artifacts in Glare repo

        :param context: user context
        :param filters: filtering conditions to Artifact list
        :param marker: id of Artifact that identifies where Glare should
        start listing Artifacts. So all Artifacts before that Artifact in
        resulting list must be ignored. It is useful for Artifact pagination.
        :param limit: maximum number of Artifact items in list.
        :param sort: sorting preferences when requesting Artifact list.
        :return: list of Artifacts
        """
        if sort is not None:
            sort = cls._parse_sort_values(sort)
        else:
            sort = [('created_at', 'desc', None), ('id', 'asc', None)]

        if filters is not None:
            filters = cls._parse_filter_values(filters)
        else:
            filters = []

        return [cls._init_artifact(context, af)
                for af in cls.db_api.list(
                context, filters, marker, limit, sort, latest)]

    @classmethod
    def delete(cls, context, af):
        """Delete Artifact and all blobs from Glare.

        :param context: user context
        :param af: definition of artifact targeted to delete
        """
        if af.visibility == 'public' and not context.is_admin:
            msg = _("Only admins are allowed to delete public artifacts")
            raise exception.Forbidden(msg)
        # marking all blobs as pending delete
        blobs = {}
        for name, field in six.iteritems(af.fields):
            if cls.is_blob(name):
                b = getattr(af, name)
                if b:
                    if b['status'] == glare_fields.\
                            BlobFieldType.PENDING_DELETE:
                        msg = _('Blob %(name)s is already deleting '
                                'for artifact %(id)s') % {'name': name,
                                                          'id': af.id}
                        raise exception.Conflict(msg)
                    else:
                        b['status'] = glare_fields.BlobFieldType.PENDING_DELETE
                        blobs[name] = b
            elif cls.is_blob_dict(name):
                bd = getattr(af, name)
                if bd:
                    for key, b in six.iteritems(bd):
                        if b['status'] == glare_fields.\
                                BlobFieldType.PENDING_DELETE:
                            msg = _('Blob %(name)s is already deleting '
                                    'for artifact %(id)s') % {'name': name,
                                                              'id': af.id}
                            raise exception.Conflict(msg)
                        else:
                            b['status'] = glare_fields.\
                                BlobFieldType.PENDING_DELETE
                    blobs[name] = bd
        if blobs:
            LOG.debug("Marked all blobs %(blobs) for artifact %(artifact)s "
                      "as pending delete. Start blobs delete.",
                      {'blobs': blobs, 'artifact': af.id})
            cls.db_api.update(context, af.id, blobs)
            # delete blobs one by one
            if not CONF.delayed_blob_delete:
                for name, blob in six.iteritems(blobs):
                    if cls.is_blob(name):
                        store_api.delete_blob(blob['url'], context=context)
                        cls.db_api.update(context, af.id, {name: None})
                    elif cls.is_blob_dict(name):
                        upd_blob = deepcopy(blob)
                        for key, val in six.iteritems(blob):
                            store_api.delete_blob(val['url'], context=context)
                            del upd_blob[key]
                            cls.db_api.update(context, af.id, {name: upd_blob})

            LOG.info(_LI("Blobs successfully deleted for artifact %s"), af.id)
        # delete artifact itself
        cls.db_api.delete(context, af.id)

    @classmethod
    def activate(cls, context, af, values):
        """Activate Artifact and make it available for users

        :param context: User Context
        :param af: current Artifact definition in Glare
        :return: definition of activated Artifact
        """
        # validate that came to artifact as updates
        if values != {'status': cls.STATUS.ACTIVE}:
            msg = _("Only {'status': %s} is allowed in a request "
                    "for activation.") % cls.STATUS.ACTIVE
            raise exception.BadRequest(msg)

        for name, type_obj in six.iteritems(af.fields):
            if type_obj.required_on_activate and getattr(af, name) is None:
                msg = _("'%s' attribute must be set before activation") % name
                raise exception.BadRequest(msg)

        cls.validate_activate(context, af)
        if af.status != cls.STATUS.DRAFTED:
            raise exception.InvalidStatusTransition(
                orig=af.status, new=cls.STATUS.ACTIVE
            )
        LOG.info(_LI("Parameters validation for artifact %(artifact)s "
                     "activate passed for request %(request)s."),
                 {'artifact': af.id, 'request': context.request_id})
        active_af = cls.db_api.update(context, af.id, values)
        return cls._init_artifact(context, active_af)

    @classmethod
    def reactivate(cls, context, af, values):
        """Make Artifact active after de-activation

        :param context: user context
        :param af: definition of de-activated Artifact
        :return: definition of active Artifact
        """
        # validate that came to artifact as updates
        if values != {'status': cls.STATUS.ACTIVE}:
            msg = _("Only {'status': %s} is allowed in a request "
                    "for reactivation.") % cls.STATUS.ACTIVE
            raise exception.BadRequest(msg)
        if af.status != cls.STATUS.DEACTIVATED:
            raise exception.InvalidStatusTransition(
                orig=af.status, new=cls.STATUS.ACTIVE
            )
        LOG.info(_LI("Parameters validation for artifact %(artifact)s "
                     "reactivate passed for request %(request)s."),
                 {'artifact': af.id, 'request': context.request_id})
        af = cls.db_api.update(context, af.id, values)
        return cls._init_artifact(context, af)

    @classmethod
    def deactivate(cls, context, af, values):
        """Deny Artifact downloading due to security concerns

        If user uploaded suspicious Artifact then Cloud Admins(or other users -
        it depends on policy configurations) can deny Artifact download by
        users by making Artifact de-activated. After additional investigation
        Artifact can be re-activated or deleted from Glare.

        :param context: user context
        :param af: Artifact definition in Glare
        :return: definition of de-activated Artifact
        """
        if values != {'status': cls.STATUS.DEACTIVATED}:
            msg = _("Only {'status': %s} is allowed in a request "
                    "for deactivation.") % cls.STATUS.DEACTIVATED
            raise exception.BadRequest(msg)

        if af.status != cls.STATUS.ACTIVE:
            raise exception.InvalidStatusTransition(
                orig=af.status, new=cls.STATUS.ACTIVE
            )
        LOG.info(_LI("Parameters validation for artifact %(artifact)s "
                     "deactivate passed for request %(request)s."),
                 {'artifact': af.id, 'request': context.request_id})
        af = cls.db_api.update(context, af.id, values)
        return cls._init_artifact(context, af)

    @classmethod
    def publish(cls, context, af, values):
        """Make Artifact available for everyone

        :param context: user context
        :param af: definition of published Artifact
        :return: definition of active Artifact
        """
        if values != {'visibility': 'public'}:
            msg = _("Only {'visibility': 'public'} is allowed in a request "
                    "for artifact publish.")
            raise exception.BadRequest(msg)

        with cls._lock_updated_version(af, values):
            if af.status != cls.STATUS.ACTIVE:
                msg = _("Cannot publish non-active artifact")
                raise exception.BadRequest(msg)

            cls._validate_versioning(context, af.name, af.version,
                                     is_public=True)
            cls.validate_publish(context, af)
            LOG.info(_LI("Parameters validation for artifact %(artifact)s "
                         "publish passed for request %(request)s."),
                     {'artifact': af.id, 'request': context.request_id})
            af = cls.db_api.update(context, af.id, values)
            return cls._init_artifact(context, af)

    @classmethod
    def _get_max_blob_size(cls, field_name):
        return getattr(cls.fields[field_name], 'max_blob_size',
                       attribute.BlobAttribute.DEFAULT_MAX_BLOB_SIZE)

    @classmethod
    def _validate_upload_allowed(cls, context, af, field_name, blob_key=None):
        if field_name not in cls.fields:
            msg = _("%s property does not exist") % field_name
            raise exception.BadRequest(msg)
        cls._validate_update_allowed(context, af, [field_name])
        if blob_key:
            if not cls.is_blob_dict(field_name):
                msg = _("%s is not a blob dict") % field_name
                raise exception.BadRequest(msg)
            if getattr(af, field_name).get(blob_key) is not None:
                msg = (_("Cannot re-upload blob value to blob dict %(blob)s "
                         "with key %(key)s for artifact %(af)s") %
                       {'blob': field_name, 'key': blob_key, 'af': af.id})
                raise exception.Conflict(message=msg)
        else:
            if not cls.is_blob(field_name):
                msg = _("%s is not a blob") % field_name
                raise exception.BadRequest(msg)
            if getattr(af, field_name) is not None:
                msg = _("Cannot re-upload blob %(blob)s for artifact "
                        "%(af)s") % {'blob': field_name, 'af': af.id}
                raise exception.Conflict(message=msg)

    @classmethod
    def upload_blob(cls, context, af, field_name, fd, content_type):
        """Upload binary object as artifact property

        :param context: user context
        :param af: current Artifact definition
        :param field_name: name of blob field
        :param fd: file descriptor that Glare uses to upload the file
        :param content_type: data content-type
        :return: updated Artifact definition in Glare
        """
        fd = cls.validate_upload(context, af, field_name, fd)
        cls._validate_upload_allowed(context, af, field_name)

        LOG.debug("Parameters validation for artifact %(artifact)s blob "
                  "upload passed for blob %(blob)s. "
                  "Start blob uploading to backend.",
                  {'artifact': af.id, 'blob': field_name})
        blob = {'url': None, 'size': None, 'md5': None, 'sha1': None,
                'sha256': None, 'status': glare_fields.BlobFieldType.SAVING,
                'external': False, 'content_type': content_type}
        setattr(af, field_name, blob)
        cls.db_api.update(
            context, af.id, {field_name: getattr(af, field_name)})
        blob_id = getattr(af, field_name)['id']

        try:
            location_uri, size, checksums = store_api.save_blob_to_store(
                blob_id, fd, context, cls._get_max_blob_size(field_name))
            blob.update({'url': location_uri,
                         'status': glare_fields.BlobFieldType.ACTIVE,
                         'size': size})
            blob.update(checksums)
            setattr(af, field_name, blob)
            af_upd = cls.db_api.update(
                context, af.id, {field_name: getattr(af, field_name)})
            LOG.info(_LI("Successfully finished blob upload for artifact "
                         "%(artifact)s blob field %(blob)s."),
                     {'artifact': af.id, 'blob': field_name})
            return cls._init_artifact(context, af_upd)
        except Exception:
            with excutils.save_and_reraise_exception(logger=LOG):
                cls.db_api.update(context, af.id, {field_name: None})

    @classmethod
    def download_blob(cls, context, af, field_name):
        """Download binary data from Glare Artifact.

        :param context: user context
        :param af: Artifact definition in Glare repo
        :param field_name: name of blob field
        :return: file iterator for requested file
        """
        if not cls.is_blob(field_name):
            msg = _("%s is not a blob") % field_name
            raise exception.BadRequest(msg)
        if af.status == cls.STATUS.DEACTIVATED and not context.is_admin:
            msg = _("Only admin is allowed to download artifact data "
                    "when it's deactivated")
            raise exception.Forbidden(message=msg)
        blob = getattr(af, field_name)
        if blob is None or blob['status'] != glare_fields.BlobFieldType.ACTIVE:
            msg = _("%s is not ready for download") % field_name
            raise exception.BadRequest(message=msg)
        meta = {'md5': blob.get('md5'),
                'sha1': blob.get('sha1'),
                'sha256': blob.get('sha256'),
                'external': blob.get('external')}
        if blob['external']:
            data = {'url': blob['url']}
        else:
            data = store_api.load_from_store(uri=blob['url'], context=context)
            meta['size'] = blob.get('size')
            meta['content_type'] = blob.get('content_type')
        return data, meta

    @classmethod
    def upload_blob_dict(cls, context, af, field_name, blob_key, fd,
                         content_type):
        """Upload binary object as artifact property

        :param context: user context
        :param af: current Artifact definition
        :param blob_key: name of blob key in dict
        :param fd: file descriptor that Glare uses to upload the file
        :param field_name: name of blob dict field
        :param content_type: data content-type
        :return: updated Artifact definition in Glare
        """
        fd = cls.validate_upload(context, af, field_name, fd)
        cls._validate_upload_allowed(context, af, field_name, blob_key)

        LOG.debug("Parameters validation for artifact %(artifact)s blob "
                  "upload passed for blob dict  %(blob)s with key %(key)s. "
                  "Start blob uploading to backend.",
                  {'artifact': af.id, 'blob': field_name, 'key': blob_key})
        blob = {'url': None, 'size': None, 'md5': None, 'sha1': None,
                'sha256': None, 'status': glare_fields.BlobFieldType.SAVING,
                'external': False, 'content_type': content_type}
        blob_dict_attr = getattr(af, field_name)
        blob_dict_attr[blob_key] = blob
        cls.db_api.update(
            context, af.id, {field_name: blob_dict_attr})
        blob_id = getattr(af, field_name)[blob_key]['id']
        try:
            location_uri, size, checksums = store_api.save_blob_to_store(
                blob_id, fd, context, cls._get_max_blob_size(field_name))
            blob.update({'url': location_uri,
                         'status': glare_fields.BlobFieldType.ACTIVE,
                         'size': size})
            blob.update(checksums)
            af_values = cls.db_api.update(
                context, af.id, {field_name: blob_dict_attr})
            LOG.info(_LI("Successfully finished blob upload for artifact "
                         "%(artifact)s blob dict field %(blob)s with key."),
                     {'artifact': af.id, 'blob': field_name, 'key': blob_key})
            return cls._init_artifact(context, af_values)
        except Exception:
            with excutils.save_and_reraise_exception(logger=LOG):
                del blob_dict_attr[blob_key]
                cls.db_api.update(context, af.id, {field_name: blob_dict_attr})

    @classmethod
    def download_blob_dict(cls, context, af, field_name, blob_key):
        """Download binary data from Glare Artifact.

        :param context: user context
        :param af: Artifact definition in Glare repo
        :param blob_key: name of blob key in dict
        :param field_name: name of blob dict field
        :return: file iterator for requested file
        """
        if not cls.is_blob_dict(field_name):
            msg = _("%s is not a blob dict") % field_name
            raise exception.BadRequest(msg)

        if af.status == cls.STATUS.DEACTIVATED and not context.is_admin:
            msg = _("Only admin is allowed to download artifact data "
                    "when it's deactivated")
            raise exception.Forbidden(message=msg)
        try:
            blob = getattr(af, field_name)[blob_key]
        except KeyError:
            msg = _("Blob with name %(blob_name)s is not found in blob "
                    "dictionary %(blob_dict)s") % (blob_key, field_name)
            raise exception.NotFound(message=msg)
        if blob is None or blob['status'] != glare_fields.BlobFieldType.ACTIVE:
            msg = _("Blob %(blob_name)s from blob dictionary %(blob_dict)s "
                    "is not ready for download") % (blob_key, field_name)
            LOG.error(msg)
            raise exception.BadRequest(message=msg)
        meta = {'md5': blob.get('md5'),
                'sha1': blob.get('sha1'),
                'sha256': blob.get('sha256'),
                'external': blob.get('external')}

        if blob['external']:
            data = {'url': blob['url']}
        else:
            data = store_api.load_from_store(uri=blob['url'], context=context)
            meta['size'] = blob.get('size')
            meta['content_type'] = blob.get('content_type')
        return data, meta

    @classmethod
    def add_blob_location(cls, context, af, field_name, location, blob_meta):
        """Upload binary object as artifact property

        :param context: user context
        :param af: current Artifact definition
        :param field_name: name of blob field
        :param location: blob url
        :return: updated Artifact definition in Glare
        """
        cls._validate_upload_allowed(context, af, field_name)
        LOG.debug("Parameters validation for artifact %(artifact)s location "
                  "passed for blob %(blob)s. Start location check for artifact"
                  ".", {'artifact': af.id, 'blob': field_name})

        blob = {'url': location, 'size': None, 'md5': None, 'sha1': None,
                'sha256': None, 'status': glare_fields.BlobFieldType.ACTIVE,
                'external': True, 'content_type': None}

        md5 = blob_meta.pop("md5", None)
        if md5 is None:
            msg = (_("Incorrect blob metadata %(meta)s. MD5 must be specified "
                     "for external location in artifact blob %(field_name)."),
                   {"meta": str(blob_meta), "field_name": field_name})
            raise exception.BadRequest(msg)
        else:
            blob["md5"] = md5
            blob["sha1"] = blob_meta.pop("sha1", None)
            blob["sha256"] = blob_meta.pop("sha256", None)

        setattr(af, field_name, blob)
        updated_af = cls.db_api.update(
            context, af.id, {field_name: getattr(af, field_name)})
        LOG.info(_LI("External location %(location)s has been created "
                     "successfully for artifact %(artifact)s blob %(blob)s"),
                 {'location': location, 'artifact': af.id,
                  'blob': field_name})
        return cls._init_artifact(context, updated_af)

    @classmethod
    def add_blob_dict_location(cls, context, af, field_name,
                               blob_key, location, blob_meta):
        cls._validate_upload_allowed(context, af, field_name, blob_key)

        blob = {'url': location, 'size': None, 'md5': None, 'sha1': None,
                'sha256': None, 'status': glare_fields.BlobFieldType.ACTIVE,
                'external': True, 'content_type': None}

        md5 = blob_meta.pop("md5", None)
        if md5 is None:
            msg = (_("Incorrect blob metadata %(meta)s. MD5 must be specified "
                     "for external location in artifact blob "
                     "%(field_name)[%(blob_key)s]."),
                   {"meta": str(blob_meta), "field_name": field_name,
                    "blob_key": str(blob_key)})
            raise exception.BadRequest(msg)
        else:
            blob["md5"] = md5
            blob["sha1"] = blob_meta.pop("sha1", None)
            blob["sha256"] = blob_meta.pop("sha256", None)

        blob_dict_attr = getattr(af, field_name)
        blob_dict_attr[blob_key] = blob
        updated_af = cls.db_api.update(
            context, af.id, {field_name: blob_dict_attr})

        LOG.info(
            _LI("External location %(location)s has been created successfully "
                "for artifact %(artifact)s blob dict %(blob)s with key "
                "%(key)s"),
            {'location': location, 'artifact': af.id,
             'blob': field_name, 'key': blob_key})
        return cls._init_artifact(context, updated_af)

    @classmethod
    def validate_activate(cls, context, af, values=None):
        pass

    @classmethod
    def validate_upload(cls, context, af, field_name, fd):
        return fd

    @classmethod
    def validate_publish(cls, context, af):
        pass

    def to_notification(self):
        """Return notification body that can be send to listeners

        :return: dict with notification information
        """
        return {
            'type': self.get_type_name(),
            'id': self.id,
            'description': self.description,
            'name': self.name,
            'version': self.version,
            'visibility': self.visibility,
            'status': self.status,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'activated_at': self.activated_at,
            'owner': self.owner
        }

    def to_dict(self):
        """Convert oslo versioned object to dictionary

        :return: dict with field names and field values
        """
        return self.obj_to_primitive()['versioned_object.data']

    def obj_changes_to_primitive(self):
        changes = self.obj_get_changes()
        res = {}
        for key, val in six.iteritems(changes):
            if val is not None and hasattr(val, 'to_primitive'):
                res[key] = val.to_primitive()
            else:
                res[key] = val
        return res

    @classmethod
    def schema_attr(cls, attr, attr_name=''):
        attr_type = utils.get_schema_type(attr)
        schema = {}

        # generate schema for validators
        for val in getattr(attr, 'validators', []):
            schema.update(val.to_jsonschema())

        schema['type'] = (attr_type
                          if not attr.nullable else [attr_type, 'null'])
        schema['glareType'] = utils.get_glare_type(attr)
        output_blob_schema = {
            'type': ['object', 'null'],
            'properties': {
                'size': {'type': ['number', 'null']},
                'md5': {'type': ['string', 'null']},
                'sha1': {'type': ['string', 'null']},
                'sha256': {'type': ['string', 'null']},
                'external': {'type': 'boolean'},
                'status': {'type': 'string',
                           'enum': list(
                               glare_fields.BlobFieldType.BLOB_STATUS)},
                'content_type': {'type': 'string'},
            },
            'required': ['size', 'md5', 'sha1', 'sha256', 'external', 'status',
                         'content_type'],
            'additionalProperties': False
        }

        if attr.system:
            schema['readOnly'] = True

        if isinstance(attr, glare_fields.Dict):
            element_type = (utils.get_schema_type(attr.element_type)
                            if hasattr(attr, 'element_type')
                            else 'string')

            if attr.element_type is glare_fields.BlobFieldType:
                schema['additionalProperties'] = output_blob_schema
            else:
                if schema.get('properties'):
                    properties = {}
                    required = schema.pop('required', [])
                    for key in schema.pop('properties'):
                        properties[key] = {
                            'type': (element_type
                                     if key in required
                                     else [element_type, 'null'])}
                    schema['properties'] = properties
                    schema['additionalProperties'] = False
                else:
                    schema['additionalProperties'] = {'type': element_type}

        if attr_type == 'array':
            schema['items'] = {
                'type': (utils.get_schema_type(attr.element_type)
                         if hasattr(attr, 'element_type')
                         else 'string')}

        if isinstance(attr, glare_fields.BlobField):
            schema.update(output_blob_schema)

        if isinstance(attr, fields.DateTimeField):
            schema['format'] = 'date-time'

        if attr_name == 'status':
            schema['enum'] = list(
                glare_fields.ArtifactStatusField.ARTIFACT_STATUS)

        if attr.description:
            schema['description'] = attr.description
        if attr.mutable:
            schema['mutable'] = True
        if attr.sortable:
            schema['sortable'] = True
        if not attr.required_on_activate:
            schema['required_on_activate'] = False
        if attr._default is not None:
            schema['default'] = attr._default

        schema['filter_ops'] = attr.filter_ops

        return schema

    @classmethod
    def gen_schemas(cls):
        schemas_prop = {}
        for attr_name, attr in six.iteritems(cls.fields):
            schemas_prop[attr_name] = cls.schema_attr(attr,
                                                      attr_name=attr_name)
        schemas = {'properties': schemas_prop,
                   'name': cls.get_type_name(),
                   'version': cls.VERSION,
                   'title': 'Artifact type %s of version %s' %
                            (cls.get_type_name(), cls.VERSION),
                   'type': 'object',
                   'required': ['name']}

        return schemas


class ReadOnlyMixin(object):
    """Mixin that disables all modifying actions on artifacts."""

    @classmethod
    def create(cls, context, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def update(cls, context, af, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def get_action_for_updates(cls, context, artifact, updates, registry):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def delete(cls, context, af):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def activate(cls, context, af, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def reactivate(cls, context, af, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def deactivate(cls, context, af, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def publish(cls, context, af, values):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def upload_blob(cls, context, af, field_name, fd, content_type):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def upload_blob_dict(cls, context, af, field_name, blob_key, fd,
                         content_type):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def add_blob_location(cls, context, af, field_name, location, blob_meta):
        raise exception.Forbidden("This type is read only.")

    @classmethod
    def add_blob_dict_location(cls, context, af, field_name,
                               blob_key, location, blob_meta):
        raise exception.Forbidden("This type is read only.")



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil

import fixtures
import glance_store as store
from glance_store import location
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_config import fixture as cfg_fixture
from oslo_db import options
from oslo_serialization import jsonutils
import testtools

from glare.common import config
from glare.common import utils

CONF = cfg.CONF


class BaseTestCase(testtools.TestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()

        self._config_fixture = self.useFixture(cfg_fixture.Config())
        config.parse_args(args=[])
        self.addCleanup(CONF.reset)
        self.test_dir = self.useFixture(fixtures.TempDir()).path
        self.conf_dir = os.path.join(self.test_dir, 'etc')
        utils.safe_mkdirs(self.conf_dir)
        self.set_policy()

    def set_policy(self):
        conf_file = "policy.json"
        self.policy_file = self._copy_data_file(conf_file, self.conf_dir)
        self.config(policy_file=self.policy_file, group='oslo_policy')

    def _copy_data_file(self, file_name, dst_dir):
        src_file_name = os.path.join('glare/tests/etc', file_name)
        shutil.copy(src_file_name, dst_dir)
        dst_file_name = os.path.join(dst_dir, file_name)
        return dst_file_name

    def set_property_protection_rules(self, rules):
        with open(self.property_file, 'w') as f:
            for rule_key in rules.keys():
                f.write('[%s]\n' % rule_key)
                for operation in rules[rule_key].keys():
                    roles_str = ','.join(rules[rule_key][operation])
                    f.write('%s = %s\n' % (operation, roles_str))

    def config(self, **kw):
        """
        Override some configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a group argument is supplied, the overrides are applied to
        the specified configuration option group.

        All overrides are automatically cleared at the end of the current
        test by the fixtures cleanup process.
        """
        self._config_fixture.config(**kw)


class StoreClearingUnitTest(BaseTestCase):

    def setUp(self):
        super(StoreClearingUnitTest, self).setUp()
        # Ensure stores + locations cleared
        location.SCHEME_TO_CLS_MAP = {}

        self._create_stores()
        self.addCleanup(setattr, location, 'SCHEME_TO_CLS_MAP', dict())

    def _create_stores(self, passing_config=True):
        """Create known stores. Mock out sheepdog's subprocess dependency
        on collie.

        :param passing_config: making store driver passes basic configurations.
        :returns: the number of how many store drivers been loaded.
        """
        store.register_opts(CONF)

        self.config(default_store='filesystem',
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")

        store.create_stores(CONF)


class IsolatedUnitTest(StoreClearingUnitTest):

    """
    Unit test case that establishes a mock environment within
    a testing directory (in isolation)
    """
    registry = None

    def setUp(self):
        super(IsolatedUnitTest, self).setUp()
        options.set_defaults(CONF, connection='sqlite:////%s/tests.sqlite' %
                                              self.test_dir)
        lockutils.set_defaults(os.path.join(self.test_dir))

        self.config(debug=False)

        self.config(default_store='filesystem',
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")

        store.create_stores()

    def set_policy_rules(self, rules):
        fap = open(CONF.oslo_policy.policy_file, 'w')
        fap.write(jsonutils.dumps(rules))
        fap.close()



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Database API for all artifact types"""

import six

from glare.db import api as base_api
from glare.db.sqlalchemy import api
from glare import locking


class ArtifactAPI(base_api.BaseDBAPI):

    def _serialize_values(self, values):
        new_values = {}
        if 'tags' in values:
            new_values['tags'] = values.pop('tags')
        for key, value in six.iteritems(values):
            if key in api.BASE_ARTIFACT_PROPERTIES:
                new_values[key] = value
            elif self.cls.is_blob(key) or self.cls.is_blob_dict(key):
                new_values.setdefault('blobs', {})[key] = value
            else:
                new_values.setdefault('properties', {})[key] = value
        return new_values

    def create(self, context, values):
        values = self._serialize_values(values)
        values['type_name'] = self.type
        session = api.get_session()
        return api.create(context, values, session)

    def update(self, context, artifact_id, values):
        session = api.get_session()
        return api.update(context, artifact_id,
                          self._serialize_values(values), session)

    def delete(self, context, artifact_id):
        session = api.get_session()
        return api.delete(context, artifact_id, session)

    def get(self, context, artifact_id):
        session = api.get_session()
        return api.get(context, artifact_id, session)

    def list(self, context, filters, marker, limit, sort, latest):
        session = api.get_session()
        if self.type != 'all':
            filters.append(('type_name', None, 'eq', None, self.type))
        return api.get_all(context=context, session=session, filters=filters,
                           marker=marker, limit=limit, sort=sort,
                           latest=latest)


class ArtifactLockApi(locking.LockApiBase):
    def create_lock(self, context, lock_key):
        session = api.get_session()
        return api.create_lock(context, lock_key, session)

    def delete_lock(self, context, lock_id):
        session = api.get_session()
        api.delete_lock(context, lock_id, session)



# Copyright 2011-2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from oslo_config import cfg
from oslo_context import context
from oslo_middleware import base as base_middleware
from oslo_middleware import request_id
from oslo_serialization import jsonutils

from glare.common import policy
from glare.i18n import _

context_opts = [
    cfg.BoolOpt('allow_anonymous_access', default=False,
                help=_('Allow unauthenticated users to access the API with '
                       'read-only privileges. This only applies when using '
                       'ContextMiddleware.'))
]

CONF = cfg.CONF
CONF.register_opts(context_opts)


class RequestContext(context.RequestContext):
    """Stores information about the security context for Glare.

    Stores how the user accesses the system, as well as additional request
    information.
    """

    def __init__(self, service_catalog=None, **kwargs):
        super(RequestContext, self).__init__(**kwargs)
        self.service_catalog = service_catalog
        # check if user is admin using policy file
        if kwargs.get('is_admin') is None:
            self.is_admin = policy.check_is_admin(self)

    def to_dict(self):
        d = super(RequestContext, self).to_dict()
        d.update({
            'service_catalog': self.service_catalog,
        })
        return d

    def to_policy_values(self):
        values = super(RequestContext, self).to_policy_values()
        values['is_admin'] = self.is_admin
        values['read_only'] = self.read_only
        return values


class ContextMiddleware(base_middleware.ConfigurableMiddleware):

    @staticmethod
    def process_request(req):
        """Convert authentication information into a request context

        Generate a RequestContext object from the available
        authentication headers and store on the 'context' attribute
        of the req object.

        :param req: wsgi request object that will be given the context object
        :raises: webob.exc.HTTPUnauthorized: when value of the
                                            X-Identity-Status  header is not
                                            'Confirmed' and anonymous access
                                            is disallowed
        """
        if req.headers.get('X-Identity-Status') == 'Confirmed':
            req.context = ContextMiddleware._get_authenticated_context(req)
        elif CONF.allow_anonymous_access:
            req.context = ContextMiddleware._get_anonymous_context()
        else:
            raise webob.exc.HTTPUnauthorized()

    @staticmethod
    def _get_anonymous_context():
        """Anonymous user has only Read-Only grants"""
        return RequestContext(read_only=True, is_admin=False)

    @staticmethod
    def _get_authenticated_context(req):
        headers = req.headers
        service_catalog = None
        if headers.get('X-Service-Catalog') is not None:
            catalog_header = headers.get('X-Service-Catalog')
            try:
                service_catalog = jsonutils.loads(catalog_header)
            except ValueError:
                raise webob.exc.HTTPInternalServerError(
                    _('Invalid service catalog json.'))
        kwargs = {
            'service_catalog': service_catalog,
            'request_id': req.environ.get(request_id.ENV_REQUEST_ID),
        }
        return RequestContext.from_environ(req.environ, **kwargs)


class UnauthenticatedContextMiddleware(base_middleware.ConfigurableMiddleware):
    """Process requests and responses when auth is turned off at all."""

    @staticmethod
    def process_request(req):
        """Create a context without an authorized user.

        When glare deployed as public repo everybody is admin
        without any credentials.
        """
        req.context = RequestContext(is_admin=True)
